/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "Brush.hh"
#include <cmath>

#include <limits>

using SharpConstruct::Brush::BrushBase;
using SharpConstruct::Brush::SupportedProperties;
using SharpConstruct::Brush::SymmetryData;
using SharpConstruct::Brush::DirectionData;
using SharpConstruct::Color;
using SharpConstruct::Math::PolarPoint;
using SharpConstruct::Brush::EditData;

SupportedProperties::SupportedProperties( const bool str, const bool iny,
					  const bool lay, const bool dot,
					  const bool msk, const bool add,
					  const bool sub, const bool col,
					  const bool pig )
:Strength( str ), Intensity( iny ), Layer( lay ), Dot( dot ), Mask( msk ),
 Add( add ), Sub( sub ), Colorize( col ), Pigment( pig )
{}

SymmetryData::SymmetryData()
: x_( false ), y_( false ), z_( false )
{}
SymmetryData::SymmetryData( const bool x, const bool y, const bool z )
: x_( x ), y_( y ), z_( z )
{}

bool SymmetryData::X() const
{
	return x_;
}

bool SymmetryData::Y() const
{
	return y_;
}

bool SymmetryData::Z() const
{
	return z_;
}

SymmetryData BrushBase::symmetry_;

BrushBase::BrushBase()
: direction_( OUT ),
  colorize_( false ),
  mask_( false ),
  radius_( 50 ),
  strength_( 25 ),
  intensity_( 25 ),
  flip_( false ),
  pressure_( 1 ),
  radius_pressure_( true ),
  strength_pressure_( true ),
  intensity_pressure_( true ),
  dot_( false ),
  layer_( false )
{}

BrushBase::~BrushBase()
{}

bool BrushBase::SingleSelect() const
{
	return false;
}

SharpConstruct::Brush::DirectionData BrushBase::Direction() const
{
	return direction_;
}
void BrushBase::SetDirection( const DirectionData d )
{
	direction_ = d;
}

bool BrushBase::Colorize() const
{
	return colorize_;
}
void BrushBase::SetColorize( const bool c )
{
	colorize_ = c;
}

bool BrushBase::Mask() const
{
	return mask_;
}
void BrushBase::SetMask( const bool m )
{
	mask_ = m;
}

bool BrushBase::AffectsVertexData() const
{
	return Direction() != OFF;
}
bool BrushBase::AffectsColorData() const
{
	return Colorize() || Mask();
}
bool BrushBase::AffectsPolygonData() const
{
	return false;
}

unsigned BrushBase::Radius() const
{
	if( Pressure() < std::numeric_limits< float >::epsilon() || !radius_pressure_ )
		return radius_;
	else
		return static_cast< unsigned >( radius_ * ( Pressure() * 0.75f + 0.25f ) );
}
unsigned BrushBase::BasicRadius() const
{
	return radius_;
}
void BrushBase::SetRadius( const unsigned r )
{
	radius_ = r;
}

float BrushBase::Strength() const
{
	float strength( strength_ );

	if( strength_pressure_ )
		strength *= Pressure();

	if( ( Direction() == OUT && Flip() ) || ( Direction() == IN && !Flip() ) )
		strength = -strength;

	return strength / factor_();
}
float BrushBase::BasicStrength() const
{
	return strength_;
}
void BrushBase::SetStrength( const float s )
{
	strength_ = s;
}

float BrushBase::Intensity() const
{
	if( intensity_pressure_ )
		return intensity_ / 100.0f * Pressure();
	else
		return intensity_ / 100.0f;
}
float BrushBase::BasicIntensity() const
{
	return intensity_;
}
void BrushBase::SetIntensity( const float i )
{
	intensity_ = i;
}

bool BrushBase::Flip() const
{
	return flip_;
}
void BrushBase::SetFlip( const bool f )
{
	flip_ = f;
}

float BrushBase::Pressure() const
{
	return pressure_;
}
void BrushBase::SetPressure( const float p )
{
	pressure_ = p;
}

bool BrushBase::RadiusPressure() const
{
	return radius_pressure_;
}
void BrushBase::SetRadiusPressure( const bool r )
{
	radius_pressure_ = r;
}

bool BrushBase::StrengthPressure() const
{
	return strength_pressure_;
}
void BrushBase::SetStrengthPressure( const bool s )
{
	strength_pressure_ = s;
}

bool BrushBase::IntensityPressure() const
{
	return intensity_pressure_;
}
void BrushBase::SetIntensityPressure( const bool i )
{
	intensity_pressure_ = i;
}

SharpConstruct::Color BrushBase::Pigment() const
{
	return pigment_;
}
void BrushBase::SetPigment( const Color p )
{
	pigment_ = p;
}

SymmetryData BrushBase::Symmetry()
{
	return symmetry_;
}
void BrushBase::SetSymmetry( const SymmetryData s )
{
	symmetry_ = s;
}

bool BrushBase::Dot() const
{
	return dot_;
}
void BrushBase::SetDot( const bool d )
{
	dot_ = d;
}

bool BrushBase::Layer() const
{
	return layer_;
}
void BrushBase::SetLayer( const bool l )
{
	layer_ = l;
}


Gtk::TreePath BrushBase::ShapePath() const
{
	return shape_path_;
}
void BrushBase::SetShape( Glib::RefPtr< Gdk::Pixbuf > s, const Gtk::TreePath p )
{
	shape_ = s;
	shape_path_ = p;
}

float BrushBase::AlphaStrength( const PolarPoint l ) const
{
	// Find alpha's center
	Math::Point2D< float > c( shape_->get_width() / 2.0f, shape_->get_height() / 2.0f );

	// Scale the magnitude to match the size of the alpha
	const float magn( l.Magnitude() * c.X() );
	
	// FIXME: NOT SURE IF THIS +C BUSINESS IS CORRECT....
	// Find the pixel in the alpha
	Math::Point2D< int > x( static_cast< int >( magn * cos( l.Angle() ) + c.X() ),
				static_cast< int >( magn * sin( l.Angle() ) + c.Y() ) );
	
	const guint8* pixels( shape_->get_pixels() );
	const guint8* p( pixels + x.Y() * shape_->get_rowstride() +
			 x.X() * shape_->get_n_channels() );

	return p[ 0 ] / 255.0f;
}

void BrushBase::StartSculpt()
{
	pre_start_sculpt_();
	symmetrical_sculpt_( *this, &BrushBase::start_sculpt_ );
	post_start_sculpt_();
}

void BrushBase::ContinueSculpt()
{
	pre_continue_sculpt_();
	symmetrical_sculpt_( *this, &BrushBase::continue_sculpt_ );
	post_continue_sculpt_();
}

float BrushBase::factor_() const
{
	return 1000;
}

void BrushBase::pre_start_sculpt_()
{
	MeshHistory::Instance().GetCurrentMesh().ClearModifiedVertexIndices();
}
void BrushBase::start_sculpt_( const EditData& e )
{}
void BrushBase::post_start_sculpt_()
{
	Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );
	mesh.RecalculateDamaged();
	mesh.RecalculateModifiedNormals();
}

void BrushBase::pre_continue_sculpt_()
{
	pre_start_sculpt_();
}
void BrushBase::continue_sculpt_( const EditData& e )
{
	start_sculpt_( e );
}
void BrushBase::post_continue_sculpt_()
{
	post_start_sculpt_();
}
