/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "ModelPreview.hh"
#include "Utilities.hh"
#include <glibmm/stringutils.h>
#include <cmath>

using SharpConstruct::ModelPreview;

ModelPreview::ModelPreview( Gtk::FileChooser& c )
: chooser_( c ),
  mesh_( vertex_data_, color_data_, polygon_data_,
	 mesh_saved_, mesh_visel_, mesh_bounds_, mesh_filename_ )
{
	set_size_request( 128, 128 );

	chooser_.signal_update_preview().connect
		( sigc::mem_fun( *this, &ModelPreview::Update ) );

	show_.add_label( "Show Preview" );
	show_.signal_toggled().connect
		( sigc::mem_fun( *this, &ModelPreview::on_show_toggled_ ) );

	padder_.pack_end( show_, Gtk::PACK_SHRINK );
	padder_.pack_end( model_display_ );
	padder_.pack_end( progress_bar_, Gtk::PACK_SHRINK );
	padder_.show();
	show_.show();
	model_display_.hide();
	add( padder_ );
}

void ModelPreview::Update()
{
	model_display_.hide();

	cancel_ = false;

	if( show_.get_active() )
	{
		const std::string file( chooser_.get_filename() );
		
		mesh_.Clear();
		
		file_.reset( new Create::File( file ) );
		chooser_.set_preview_widget_active( file_->Load( mesh_ ) );
		
		finalize_creation_.reset( new FinalizeCreation( mesh_ ) );
		
		Glib::signal_idle().connect( sigc::mem_fun( *this, &ModelPreview::step_ ) );
		
		progress_bar_.show();
	}
	else
		progress_bar_.hide();
}

bool ModelPreview::step_()
{
	if( cancel_ )
		return false;
	if( file_->Step() )
	{
		update_progress_bar_( file_->Progress(),
				      file_->Text() );
		return true;
	}
	else if( finalize_creation_->Step() )
	{
		update_progress_bar_( finalize_creation_->Progress(),
				      finalize_creation_->Text() );
		return true;
	}
	else
	{
		model_display_.show();
		progress_bar_.hide();
		model_display_.SetMesh( mesh_ );
		return false;
	}
}

void ModelPreview::update_progress_bar_( const double p,
					 const std::string& txt )
{
		static double PrevTime( Time() );
		
		if( Time() > PrevTime + 0.1 )
		{
			double pc( p );
			if( pc < 0 )
				pc = 0;
			else if( pc > 1 )
				pc = 1;
			progress_bar_.set_text( txt );
			progress_bar_.set_fraction( pc );
			PrevTime = Time();
		}
}

void ModelPreview::on_show_toggled_()
{
	if( show_.get_active() )
	{
		Update();
	}
	else
	{
		cancel_ = true;
		model_display_.hide();
		progress_bar_.hide();
	}
}
